# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""
Readers for lvl1b and lvl2 data in eps format.
"""
from __future__ import division

import os
import fnmatch
import lxml.etree as etree
from tempfile import NamedTemporaryFile
from gzip import GzipFile
from collections import OrderedDict
from pygeobase.io_base import ImageBase
from pygeobase.object_base import Image

import numpy as np

short_cds_time = np.dtype([('day', np.uint16), ('time', np.uint32)])

long_cds_time = np.dtype([('day', np.uint16), ('ms', np.uint32),
                          ('mms', np.uint16)])

long_nan = np.iinfo(np.int32).min
ulong_nan = np.iinfo(np.uint32).max
int_nan = np.iinfo(np.int16).min
uint_nan = np.iinfo(np.uint16).max
byte_nan = np.iinfo(np.byte).min

# 1.1.2000 00:00:00 as jd
julian_epoch = 2451544.5


class AscatL1bEPSImage(ImageBase):
    def __init__(self, *args, **kwargs):
        """
        Initialization of i/o object.

        SZF - beam_num
        - 1 Left Fore Antenna
        - 2 Left Mid Antenna
        - 3 Left Aft Antenna
        - 4 Right Fore Antenna
        - 5 Right Mid Antenna
        - 6 Right Aft Antenna
        """
        super(AscatL1bEPSImage, self).__init__(*args, **kwargs)

    def read(self, timestamp=None, file_format=None, **kwargs):
        img = read_eps_l1b(
            self.filename, timestamp)

        return img

    def write(self, *args, **kwargs):
        pass

    def flush(self):
        pass

    def close(self):
        pass


class AscatL2EPSImage(ImageBase):
    def __init__(self, *args, **kwargs):
        """
        Initialization of i/o object.
        """
        super(AscatL2EPSImage, self).__init__(*args, **kwargs)

    def read(self, timestamp=None, file_format=None, **kwargs):
        img = read_eps_l2(self.filename, timestamp)

        return img

    def write(self, *args, **kwargs):
        pass

    def flush(self):
        pass

    def close(self):
        pass


class EPSProduct(object):
    """
    Class for reading EPS products.
    """

    def __init__(self, filename):
        self.filename = filename
        self.fid = None
        self.grh = None
        self.mphr = None
        self.sphr = None
        self.ipr = None
        self.geadr = None
        self.giadr_archive = None
        self.veadr = None
        self.viadr = None
        self.viadr_scaled = None
        self.viadr_grid = None
        self.viadr_grid_scaled = None
        self.dummy_mdr = None
        self.mdr = None
        self.eor = 0
        self.bor = 0
        self.mdr_counter = 0
        self.filesize = 0
        self.xml_file = None
        self.xml_doc = None
        self.mdr_template = None
        self.scaled_mdr = None
        self.scaled_template = None
        self.sfactor = None

    def read_product(self):
        """
        Read complete file and create numpy arrays from raw byte string data.
        """
        # open file in read-binary mode
        self.fid = open(self.filename, 'rb')
        self.filesize = os.path.getsize(self.filename)
        self.eor = self.fid.tell()

        # loop as long as the current position hasn't reached the end of file
        while self.eor < self.filesize:

            # remember beginning of the record
            self.bor = self.fid.tell()

            # read grh of current record (Generic Record Header)
            self.grh = self._read_record(grh_record())
            record_size = self.grh[0]['record_size']
            record_class = self.grh[0]['record_class']
            record_subclass = self.grh[0]['record_subclass']

            # mphr (Main Product Header Reader)
            if record_class == 1:
                self._read_mphr()

                # find the xml file corresponding to the format version
                self.xml_file = self._get_eps_xml()
                self.xml_doc = etree.parse(self.xml_file)
                self.mdr_template, self.scaled_template, self.sfactor = \
                    self._read_xml_mdr()

            # sphr (Secondary Product Header Record)
            elif record_class == 2:
                self._read_sphr()

            # ipr (Internal Pointer Record)
            elif record_class == 3:
                ipr_element = self._read_record(ipr_record())
                if self.ipr is None:
                    self.ipr = {}
                    self.ipr['data'] = [ipr_element]
                    self.ipr['grh'] = [self.grh]
                else:
                    self.ipr['data'].append(ipr_element)
                    self.ipr['grh'].append(self.grh)

            # geadr (Global External Auxiliary Data Record)
            elif record_class == 4:
                geadr_element = self._read_pointer()
                if self.geadr is None:
                    self.geadr = {}
                    self.geadr['data'] = [geadr_element]
                    self.geadr['grh'] = [self.grh]
                else:
                    self.geadr['data'].append(geadr_element)
                    self.geadr['grh'].append(self.grh)

            # veadr (Variable External Auxiliary Data Record)
            elif record_class == 6:
                veadr_element = self._read_pointer()
                if self.veadr is None:
                    self.veadr = {}
                    self.veadr['data'] = [veadr_element]
                    self.veadr['grh'] = [self.grh]
                else:
                    self.veadr['data'].append(veadr_element)
                    self.veadr['grh'].append(self.grh)

            # viadr (Variable Internal Auxiliary Data Record)
            elif record_class == 7:
                template, scaled_template, sfactor = self._read_xml_viadr(
                    record_subclass)
                viadr_element = self._read_record(template)
                viadr_element_sc = self._scaling(viadr_element,
                                                 scaled_template, sfactor)

                # store viadr_grid separately
                if record_subclass == 8:
                    if self.viadr_grid is None:
                        self.viadr_grid = [viadr_element]
                        self.viadr_grid_scaled = [viadr_element_sc]
                    else:
                        self.viadr_grid.append(viadr_element)
                        self.viadr_grid_scaled.append(viadr_element_sc)
                else:
                    if self.viadr is None:
                        self.viadr = {}
                        self.viadr_scaled = {}
                        self.viadr['data'] = [viadr_element]
                        self.viadr['grh'] = [self.grh]
                        self.viadr_scaled['data'] = [viadr_element_sc]
                        self.viadr_scaled['grh'] = [self.grh]
                    else:
                        self.viadr['data'].append(viadr_element)
                        self.viadr['grh'].append(self.grh)
                        self.viadr_scaled['data'].append(viadr_element_sc)
                        self.viadr_scaled['grh'].append(self.grh)

            # mdr (Measurement Data Record)
            elif record_class == 8:
                if self.grh[0]['instrument_group'] == 13:
                    self.dummy_mdr = self._read_record(self.mdr_template)
                else:
                    mdr_element = self._read_record(self.mdr_template)
                    if self.mdr is None:
                        self.mdr = [mdr_element]
                    else:
                        self.mdr.append(mdr_element)
                    self.mdr_counter += 1

            else:
                raise RuntimeError("Record class not found.")

            # return pointer to the beginning of the record
            self.fid.seek(self.bor)
            self.fid.seek(record_size, 1)

            # determine number of bytes read
            # end of record
            self.eor = self.fid.tell()

        self.fid.close()

        self.mdr = np.hstack(self.mdr)
        self.scaled_mdr = self._scaling(self.mdr, self.scaled_template,
                                        self.sfactor)

    def _append_list(self, new_element):
        return

    def _scaling(self, unscaled_data, scaled_template, sfactor):
        """
        Scale the data
        """
        scaled_data = np.zeros_like(unscaled_data, dtype=scaled_template)

        for name, sf in zip(unscaled_data.dtype.names, sfactor):
            if sf != 1:
                scaled_data[name] = unscaled_data[name] / sf
            else:
                scaled_data[name] = unscaled_data[name]

        return scaled_data

    def _read_record(self, dtype, count=1):
        """
        Read record
        """
        record = np.fromfile(self.fid, dtype=dtype, count=count)
        return record.newbyteorder('B')

    def _read_mphr(self):
        """
        Read Main Product Header (MPHR).
        """
        mphr = self.fid.read(self.grh[0]['record_size'] - self.grh[0].itemsize)
        self.mphr = OrderedDict(item.replace(' ', '').split('=')
                                for item in
                                mphr.decode("utf-8").split('\n')[:-1])

    def _read_sphr(self):
        """
        Read Special Product Header (SPHR).
        """
        sphr = self.fid.read(self.grh[0]['record_size'] - self.grh[0].itemsize)
        self.sphr = OrderedDict(item.replace(' ', '').split('=')
                                for item in
                                sphr.decode("utf-8").split('\n')[:-1])

    def _read_pointer(self, count=1):
        """
        Read pointer record.
        """
        dtype = np.dtype([('aux_data_pointer', np.ubyte, 100)])
        record = np.fromfile(self.fid, dtype=dtype, count=count)
        return record.newbyteorder('B')

    def _get_eps_xml(self):
        """
        Find the corresponding eps xml file.
        """
        format_path = os.path.join(os.path.dirname(__file__), '..',
                                   'formats')

        # loop through files where filename starts with 'eps_ascat'.
        for filename in fnmatch.filter(os.listdir(format_path), 'eps_ascat*'):
            doc = etree.parse(os.path.join(format_path, filename))
            file_extension = doc.xpath('//file-extensions')[0].getchildren()[0]

            format_version = doc.xpath('//format-version')
            for elem in format_version:
                major = elem.getchildren()[0]
                minor = elem.getchildren()[1]

                # return the xml file matching the metadata of the datafile.
                if major.text == self.mphr['FORMAT_MAJOR_VERSION'] and \
                        minor.text == self.mphr['FORMAT_MINOR_VERSION'] and \
                        self.mphr[
                            'PROCESSING_LEVEL'] in file_extension.text and \
                        self.mphr['PRODUCT_TYPE'] in file_extension.text:
                    return os.path.join(format_path, filename)

    def _read_xml_viadr(self, subclassid):
        """
        Read xml record of viadr class.
        """
        elements = self.xml_doc.xpath('//viadr')
        data = OrderedDict()
        length = []

        # find the element with the correct subclass
        for elem in elements:
            item_dict = dict(elem.items())
            subclass = int(item_dict['subclass'])
            if subclass == subclassid:
                break

        for child in elem.getchildren():

            if child.tag == 'delimiter':
                continue

            child_items = dict(child.items())
            name = child_items.pop('name')

            # check if the item is of type longtime
            longtime_flag = ('type' in child_items and
                             'longtime' in child_items['type'])

            # append the length if it isn't the special case of type longtime
            try:
                var_len = child_items.pop('length')
                if not longtime_flag:
                    length.append(np.int(var_len))
            except KeyError:
                pass

            data[name] = child_items

            if child.tag == 'array':
                for arr in child.iterdescendants():
                    arr_items = dict(arr.items())
                    if arr.tag == 'field':
                        data[name].update(arr_items)
                    else:
                        try:
                            var_len = arr_items.pop('length')
                            length.append(np.int(var_len))
                        except KeyError:
                            pass

            if length:
                data[name].update({'length': length})
            else:
                data[name].update({'length': 1})

            length = []

        conv = {'longtime': long_cds_time, 'time': short_cds_time,
                'boolean': np.uint8, 'integer1': np.int8,
                'uinteger1': np.uint8, 'integer': np.int32,
                'uinteger': np.uint32, 'integer2': np.int16,
                'uinteger2': np.uint16, 'integer4': np.int32,
                'uinteger4': np.uint32, 'integer8': np.int64,
                'enumerated': np.uint8, 'string': 'str', 'bitfield': np.uint8}

        scaling_factor = []
        scaled_dtype = []
        dtype = []

        for key, value in data.items():

            if 'scaling-factor' in value:
                sf_dtype = np.float32
                sf = float(eval(value['scaling-factor'].replace('^', '**')))
            else:
                sf_dtype = conv[value['type']]
                sf = 1.

            if not isinstance(value['length'], list):
                length = [value['length']]
            else:
                length = value['length']

            scaling_factor.append(sf)
            scaled_dtype.append((key, sf_dtype, length))
            dtype.append((key, conv[value['type']], length))

        return np.dtype(dtype), np.dtype(scaled_dtype), np.array(scaling_factor)

    def _read_xml_mdr(self):
        """
        Read xml record of mdr class.
        """
        elements = self.xml_doc.xpath('//mdr')
        data = OrderedDict()
        length = []
        elem = elements[0]

        for child in elem.getchildren():

            if child.tag == 'delimiter':
                continue

            child_items = dict(child.items())
            name = child_items.pop('name')

            # check if the item is of type bitfield
            bitfield_flag = ('type' in child_items and
                             ('bitfield' in child_items['type'] or 'time' in
                              child_items['type']))

            # append the length if it isn't the special case of type
            # bitfield or time
            try:
                var_len = child_items.pop('length')
                if not bitfield_flag:
                    length.append(np.int(var_len))
            except KeyError:
                pass

            data[name] = child_items

            if child.tag == 'array':
                for arr in child.iterdescendants():
                    arr_items = dict(arr.items())

                    # check if the type is bitfield
                    bitfield_flag = ('type' in arr_items and
                                     'bitfield' in arr_items['type'])

                    if bitfield_flag:
                        data[name].update(arr_items)
                        break
                    else:
                        if arr.tag == 'field':
                            data[name].update(arr_items)
                        else:
                            try:
                                var_len = arr_items.pop('length')
                                length.append(np.int(var_len))
                            except KeyError:
                                pass

            if length:
                data[name].update({'length': length})
            else:
                data[name].update({'length': 1})

            length = []

        conv = {'longtime': long_cds_time, 'time': short_cds_time,
                'boolean': np.uint8, 'integer1': np.int8,
                'uinteger1': np.uint8, 'integer': np.int32,
                'uinteger': np.uint32, 'integer2': np.int16,
                'uinteger2': np.uint16, 'integer4': np.int32,
                'uinteger4': np.uint32, 'integer8': np.int64,
                'enumerated': np.uint8, 'string': 'str', 'bitfield': np.uint8}

        scaling_factor = []
        scaled_dtype = []
        dtype = []

        for key, value in data.items():

            if 'scaling-factor' in value:
                sf_dtype = np.float32
                sf = float(eval(value['scaling-factor'].replace('^', '**')))
            else:
                sf_dtype = conv[value['type']]
                sf = 1.

            scaling_factor.append(sf)

            if not isinstance(value['length'], list):
                length = [value['length']]
            else:
                length = value['length']

            scaled_dtype.append((key, sf_dtype, length))
            dtype.append((key, conv[value['type']], length))

        return np.dtype(dtype), np.dtype(scaled_dtype), np.array(scaling_factor)


def grh_record():
    """
    Generic record header.
    """
    record_dtype = np.dtype([('record_class', np.ubyte),
                             ('instrument_group', np.ubyte),
                             ('record_subclass', np.ubyte),
                             ('record_subclass_version', np.ubyte),
                             ('record_size', np.uint32),
                             ('record_start_time', short_cds_time),
                             ('record_stop_time', short_cds_time)])

    return record_dtype


def ipr_record():
    """
    ipr template.
    """
    record_dtype = np.dtype([('target_record_class', np.ubyte),
                             ('target_instrument_group', np.ubyte),
                             ('target_record_subclass', np.ubyte),
                             ('target_record_offset', np.uint32)])
    return record_dtype


def read_eps_l1b(filename, timestamp):
    """
    Use of correct lvl1b reader and data preparation.
    """
    data = {}
    eps_file = read_eps(filename)
    ptype = eps_file.mphr['PRODUCT_TYPE']
    fmv = int(eps_file.mphr['FORMAT_MAJOR_VERSION'])

    if ptype == 'SZF':
        image_dict = {'img1': {}, 'img2': {}, 'img3': {}, 'img4': {},
                      'img5': {}, 'img6': {}}
        data_full = {'d1': {}, 'd2': {}, 'd3': {}, 'd4': {}, 'd5': {},
                     'd6': {}}
        if fmv == 12:
            raw_data, metadata, orbit_grid = read_szf_fmv_12(eps_file)
        else:
            raise RuntimeError("SZF format version not supported.")

        for field in raw_data:
            data[field] = raw_data[field]

        # separate data into single beam images
        for i in range(1, 7):
            dataset = 'd' + str(i)
            img = 'img' + str(i)
            mask = ((data['BEAM_NUMBER']) == i)
            for field in data:
                data_full[dataset][field] = data[field][mask]

            lon = data_full[dataset].pop('LONGITUDE_FULL')
            lat = data_full[dataset].pop('LATITUDE_FULL')
            image_dict[img] = Image(lon, lat, data_full[dataset], metadata,
                                    timestamp, timekey='jd')

        return image_dict

    elif (ptype == 'SZR') or (ptype == 'SZO'):
        if fmv == 11:
            raw_data, metadata = read_szx_fmv_11(eps_file)
        elif fmv == 12:
            raw_data, metadata = read_szx_fmv_12(eps_file)
        else:
            raise RuntimeError("SZX format version not supported.")

        beams = ['f_', 'm_', 'a_']

        for field in raw_data:
            if len(raw_data[field].shape) == 1:
                data[field] = raw_data[field]
            # split data if it is triplet data
            elif len(raw_data[field].shape) == 2:
                for i, beam in enumerate(beams):
                    data[beam + field] = raw_data[field][:, i]
            else:
                raise RuntimeError("Unexpected variable shape.")

        longitude = data.pop('LONGITUDE')
        latitude = data.pop('LATITUDE')

        return Image(longitude, latitude, data, metadata, timestamp,
                     timekey='jd')
    else:
        raise ValueError("Format not supported. Product type {:1}"
                         " Format major version: {:2}".format(ptype, fmv))


def read_eps_l2(filename, timestamp):
    """
    Use of correct lvl2 reader and data preparation.
    """
    data = {}
    eps_file = read_eps(filename)
    ptype = eps_file.mphr['PRODUCT_TYPE']
    fmv = int(eps_file.mphr['FORMAT_MAJOR_VERSION'])

    if (ptype == 'SMR') or (ptype == 'SMO'):
        if fmv == 12:
            raw_data, metadata = read_smx_fmv_12(eps_file)
        else:
            raise RuntimeError("SMX format version not supported.")

        beams = ['f_', 'm_', 'a_']

        for field in raw_data:
            if len(raw_data[field].shape) == 1:
                data[field] = raw_data[field]
            elif len(raw_data[field].shape) == 2:
                for i, beam in enumerate(beams):
                    data[beam + field] = raw_data[field][:, i]
            else:
                raise RuntimeError("Unexpected variable shape.")

        longitude = data.pop('LONGITUDE')
        latitude = data.pop('LATITUDE')

    else:
        raise ValueError("Format not supported. Product type {:1}"
                         " Format major version: {:2}".format(ptype, fmv))

    return Image(longitude, latitude, data, metadata, timestamp, timekey='jd')


def read_eps(filename):
    """
    Read EPS file.
    """
    zipped = False
    if os.path.splitext(filename)[1] == '.gz':
        zipped = True

    # for zipped files use an unzipped temporary copy
    if zipped:
        with NamedTemporaryFile(delete=False) as tmp_fid:
            with GzipFile(filename) as gz_fid:
                tmp_fid.write(gz_fid.read())
            filename = tmp_fid.name

    # create the eps object with the filename and read it
    prod = EPSProduct(filename)
    prod.read_product()

    # remove the temporary copy
    if zipped:
        os.remove(filename)

    return prod


def read_szx_fmv_11(eps_file):
    """
    Read SZO/SZR format version 11.

    Parameters
    ----------
    eps_file : EPSProduct object
        EPS Product object.

    Returns
    -------
    data : numpy.ndarray
        SZO/SZR data.
    """
    raw_data = eps_file.scaled_mdr
    raw_unscaled = eps_file.mdr
    mphr = eps_file.mphr

    n_node_per_line = raw_data['LONGITUDE'].shape[1]
    n_lines = raw_data['LONGITUDE'].shape[0]
    n_records = raw_data['LONGITUDE'].size
    data = {}
    metadata = {}
    idx_nodes = np.arange(n_lines).repeat(n_node_per_line)

    ascat_time = shortcdstime2jd(raw_data['UTC_LINE_NODES'].flatten()['day'],
                                 raw_data['UTC_LINE_NODES'].flatten()['time'])
    data['jd'] = ascat_time[idx_nodes]

    metadata['SPACECRAFT_ID'] = np.int8(mphr['SPACECRAFT_ID'][-1])
    metadata['ORBIT_START'] = np.uint32(mphr['ORBIT_START'])

    fields = ['PROCESSOR_MAJOR_VERSION', 'PROCESSOR_MINOR_VERSION',
              'FORMAT_MAJOR_VERSION', 'FORMAT_MINOR_VERSION']
    for field in fields:
        # metadata[field] = np.repeat(np.int16(mphr[field]),n_records)
        metadata[field] = np.int16(mphr[field])

    fields = ['SAT_TRACK_AZI']
    for field in fields:
        data[field] = raw_data[field].flatten()[idx_nodes]

    fields = [('LONGITUDE', long_nan),
              ('LATITUDE', long_nan),
              ('SWATH_INDICATOR', byte_nan)]
    for field in fields:
        data[field[0]] = raw_data[field[0]].flatten()
        valid = raw_unscaled[field[0]].flatten() != field[1]
        data[field[0]][~valid] = field[1]

    fields = [('SIGMA0_TRIP', long_nan),
              ('INC_ANGLE_TRIP', uint_nan),
              ('AZI_ANGLE_TRIP', int_nan),
              ('KP', uint_nan),
              ('F_KP', byte_nan),
              ('F_USABLE', byte_nan),
              ('F_F', uint_nan),
              ('F_V', uint_nan),
              ('F_OA', uint_nan),
              ('F_SA', uint_nan),
              ('F_TEL', uint_nan),
              ('F_LAND', uint_nan)]
    for field in fields:
        data[field[0]] = raw_data[field[0]].reshape(n_records, 3)
        # valid = data[field[0]] != field[2]
        valid = raw_unscaled[field[0]].reshape(n_records, 3) != field[1]
        data[field[0]][~valid] = field[1]

    # modify longitudes from (0, 360) to (-180,180)
    mask = np.logical_and(data['LONGITUDE'] != long_nan,
                          data['LONGITUDE'] > 180)
    data['LONGITUDE'][mask] += -360.

    # modify azimuth from (-180, 180) to (0, 360)
    mask = (data['AZI_ANGLE_TRIP'] != int_nan) & (data['AZI_ANGLE_TRIP'] < 0)
    data['AZI_ANGLE_TRIP'][mask] += 360

    data['NODE_NUM'] = np.tile((np.arange(n_node_per_line) + 1),
                               n_lines)

    data['LINE_NUM'] = idx_nodes

    data['AS_DES_PASS'] = (data['SAT_TRACK_AZI'] < 270).astype(np.uint8)

    data['SWATH INDICATOR'] = data.pop('SWATH_INDICATOR')

    return data, metadata


def read_szx_fmv_12(eps_file):
    """
    Read SZO/SZR format version 12.

    Parameters
    ----------
    eps_file : EPSProduct object
        EPS Product object.

    Returns
    -------
    data : numpy.ndarray
        SZO/SZR data.
    """
    raw_data = eps_file.scaled_mdr
    raw_unscaled = eps_file.mdr
    mphr = eps_file.mphr

    n_node_per_line = raw_data['LONGITUDE'].shape[1]
    n_lines = raw_data['LONGITUDE'].shape[0]
    n_records = raw_data['LONGITUDE'].size
    data = {}
    metadata = {}
    idx_nodes = np.arange(n_lines).repeat(n_node_per_line)

    ascat_time = shortcdstime2jd(raw_data['UTC_LINE_NODES'].flatten()['day'],
                                 raw_data['UTC_LINE_NODES'].flatten()['time'])
    data['jd'] = ascat_time[idx_nodes]

    metadata['SPACECRAFT_ID'] = np.int8(mphr['SPACECRAFT_ID'][-1])
    metadata['ORBIT_START'] = np.uint32(mphr['ORBIT_START'])

    fields = ['PROCESSOR_MAJOR_VERSION', 'PROCESSOR_MINOR_VERSION',
              'FORMAT_MAJOR_VERSION', 'FORMAT_MINOR_VERSION']
    for field in fields:
        # metadata[field] = np.repeat(np.int16(mphr[field]),n_records)
        metadata[field] = np.int16(mphr[field])

    fields = ['DEGRADED_INST_MDR', 'DEGRADED_PROC_MDR', 'SAT_TRACK_AZI',
              'ABS_LINE_NUMBER']
    for field in fields:
        data[field] = raw_data[field].flatten()[idx_nodes]

    fields = [('LONGITUDE', long_nan),
              ('LATITUDE', long_nan),
              ('SWATH INDICATOR', byte_nan)]
    for field in fields:
        data[field[0]] = raw_data[field[0]].flatten()
        valid = raw_unscaled[field[0]].flatten() != field[1]
        data[field[0]][~valid] = field[1]

    fields = [('SIGMA0_TRIP', long_nan),
              ('INC_ANGLE_TRIP', uint_nan),
              ('AZI_ANGLE_TRIP', int_nan),
              ('KP', uint_nan),
              ('NUM_VAL_TRIP', ulong_nan),
              ('F_KP', byte_nan),
              ('F_USABLE', byte_nan),
              ('F_F', uint_nan),
              ('F_V', uint_nan),
              ('F_OA', uint_nan),
              ('F_SA', uint_nan),
              ('F_TEL', uint_nan),
              ('F_REF', uint_nan),
              ('F_LAND', uint_nan)]
    for field in fields:
        data[field[0]] = raw_data[field[0]].reshape(n_records, 3)
        # valid = data[field[0]] != field[2]
        valid = raw_unscaled[field[0]].reshape(n_records, 3) != field[1]
        data[field[0]][~valid] = field[1]

    # modify longitudes from (0, 360) to (-180,180)
    mask = np.logical_and(data['LONGITUDE'] != long_nan,
                          data['LONGITUDE'] > 180)
    data['LONGITUDE'][mask] += -360.

    # modify azimuth from (-180, 180) to (0, 360)
    mask = (data['AZI_ANGLE_TRIP'] != int_nan) & (data['AZI_ANGLE_TRIP'] < 0)
    data['AZI_ANGLE_TRIP'][mask] += 360

    data['NODE_NUM'] = np.tile((np.arange(n_node_per_line) + 1),
                               n_lines)

    data['LINE_NUM'] = idx_nodes

    data['AS_DES_PASS'] = (data['SAT_TRACK_AZI'] < 270).astype(np.uint8)

    return data, metadata


def read_szf_fmv_12(eps_file):
    """
    Read SZF format version 12.

    beam_num
    - 1 Left Fore Antenna
    - 2 Left Mid Antenna
    - 3 Left Aft Antenna
    - 4 Right Fore Antenna
    - 5 Right Mid Antenna
    - 6 Right Aft Antenna

    as_des_pass
    - 0 Ascending
    - 1 Descending

    swath_indicator
    - 0 Left
    - 1 Right

    Parameters
    ----------
    eps_file : EPSProduct object
        EPS Product object.

    Returns
    -------
    data : numpy.ndarray
        SZF data.
    orbit_gri : numpy.ndarray
        6.25km orbit lat/lon grid.
    """
    raw_data = eps_file.scaled_mdr
    mphr = eps_file.mphr

    n_node_per_line = raw_data['LONGITUDE_FULL'].shape[1]
    n_lines = eps_file.mdr_counter
    data = {}
    metadata = {}
    idx_nodes = np.arange(n_lines).repeat(n_node_per_line)

    ascat_time = shortcdstime2jd(raw_data['UTC_LOCALISATION'].flatten()['day'],
                                 raw_data['UTC_LOCALISATION'].flatten()[
                                     'time'])
    data['jd'] = ascat_time[idx_nodes]

    metadata['SPACECRAFT_ID'] = np.int8(mphr['SPACECRAFT_ID'][-1])
    metadata['ORBIT_START'] = np.uint32(eps_file.mphr['ORBIT_START'])

    fields = ['PROCESSOR_MAJOR_VERSION', 'PROCESSOR_MINOR_VERSION',
              'FORMAT_MAJOR_VERSION', 'FORMAT_MINOR_VERSION']
    for field in fields:
        metadata[field] = np.int16(mphr[field])

    fields = ['DEGRADED_INST_MDR', 'DEGRADED_PROC_MDR', 'SAT_TRACK_AZI',
              'BEAM_NUMBER', 'FLAGFIELD_RF1', 'FLAGFIELD_RF2',
              'FLAGFIELD_PL', 'FLAGFIELD_GEN1']
    for field in fields:
        data[field] = raw_data[field].flatten()[idx_nodes]

    data['SWATH_INDICATOR'] = np.int8(data['BEAM_NUMBER'].flatten() > 3)

    fields = [('LONGITUDE_FULL', long_nan),
              ('LATITUDE_FULL', long_nan),
              ('SIGMA0_FULL', long_nan),
              ('INC_ANGLE_FULL', uint_nan),
              ('AZI_ANGLE_FULL', int_nan),
              ('LAND_FRAC', uint_nan),
              ('FLAGFIELD_GEN2', byte_nan)]
    for field in fields:
        data[field[0]] = raw_data[field[0]].flatten()
        valid = data[field[0]] != field[1]
        data[field[0]][valid] = data[field[0]][valid]

    # modify longitudes from (0, 360) to (-180,180)
    mask = np.logical_and(data['LONGITUDE_FULL'] != long_nan,
                          data['LONGITUDE_FULL'] > 180)
    data['LONGITUDE_FULL'][mask] += -360.

    # modify azimuth from (-180, 180) to (0, 360)
    mask = (data['AZI_ANGLE_FULL'] != int_nan) & (data['AZI_ANGLE_FULL'] < 0)
    data['AZI_ANGLE_FULL'][mask] += 360

    grid_nodes_per_line = 2 * 81

    viadr_grid = np.concatenate(eps_file.viadr_grid)
    orbit_grid = np.zeros(viadr_grid.size * grid_nodes_per_line,
                          dtype=np.dtype([('lon', np.float32),
                                          ('lat', np.float32),
                                          ('node_num', np.int16),
                                          ('line_num', np.int32)]))

    for pos_all in range(orbit_grid['lon'].size):
        line = pos_all // grid_nodes_per_line
        pos_small = pos_all % 81
        if pos_all % grid_nodes_per_line <= 80:
            # left swath
            orbit_grid['lon'][pos_all] = viadr_grid[
                'LONGITUDE_LEFT'][line][80 - pos_small]
            orbit_grid['lat'][pos_all] = viadr_grid[
                'LATITUDE_LEFT'][line][80 - pos_small]
        else:
            # right swath
            orbit_grid['lon'][pos_all] = viadr_grid[
                'LONGITUDE_RIGHT'][line][pos_small]
            orbit_grid['lat'][pos_all] = viadr_grid[
                'LATITUDE_RIGHT'][line][pos_small]

    orbit_grid['node_num'] = np.tile((np.arange(grid_nodes_per_line) + 1),
                                     viadr_grid.size)

    lines = np.arange(0, viadr_grid.size * 2, 2)
    orbit_grid['line_num'] = np.repeat(lines, grid_nodes_per_line)

    fields = ['lon', 'lat']
    for field in fields:
        orbit_grid[field] = orbit_grid[field] * 1e-6

    mask = (orbit_grid['lon'] != long_nan) & (orbit_grid['lon'] > 180)
    orbit_grid['lon'][mask] += -360.

    set_flags(data)

    data['AS_DES_PASS'] = (data['SAT_TRACK_AZI'] < 270).astype(np.uint8)

    return data, metadata, orbit_grid


def set_flags(data):
    """
    Compute summary flag for each measurement with a value of 0, 1 or 2
    indicating nominal, slightly degraded or severely degraded data.

    Parameters
    ----------
    data : numpy.ndarray
        SZF data.
    """

    # category:status = 'red': 2, 'amber': 1, 'warning': 0
    flag_status_bit = {'FLAGFIELD_RF1': {'2': [2, 4],
                                         '1': [0, 1, 3]},

                       'FLAGFIELD_RF2': {'2': [0, 1]},

                       'FLAGFIELD_PL': {'2': [0, 1, 2, 3],
                                        '0': [4]},

                       'FLAGFIELD_GEN1': {'2': [1],
                                          '0': [0]},

                       'FLAGFIELD_GEN2': {'2': [2],
                                          '1': [0],
                                          '0': [1]}
                       }

    for flagfield in flag_status_bit.keys():
        # get flag data in binary format to get flags
        unpacked_bits = np.unpackbits(data[flagfield])

        # find indizes where a flag is set
        set_bits = np.where(unpacked_bits == 1)[0]
        if set_bits.size != 0:
            pos_8 = 7 - (set_bits % 8)

            for category in sorted(flag_status_bit[flagfield].keys()):
                if (int(category) == 0) and (flagfield != 'FLAGFIELD_GEN2'):
                    continue

                for bit2check in flag_status_bit[flagfield][category]:
                    pos = np.where(pos_8 == bit2check)[0]
                    data['F_USABLE'] = np.zeros(data['FLAGFIELD_GEN2'].size)
                    data['F_USABLE'][set_bits[pos] // 8] = int(category)

                    # land points
                    if (flagfield == 'FLAGFIELD_GEN2') and (bit2check == 1):
                        data['F_LAND'] = np.zeros(data['FLAGFIELD_GEN2'].size)
                        data['F_LAND'][set_bits[pos] // 8] = 1


def read_smx_fmv_12(eps_file):
    """
    Read SMO/SMR format version 12.

    Parameters
    ----------
    eps_file : EPSProduct object
        EPS Product object.

    Returns
    -------
    data : numpy.ndarray
        SMO/SMR data.
    """
    raw_data = eps_file.scaled_mdr
    raw_unscaled = eps_file.mdr

    n_node_per_line = raw_data['LONGITUDE'].shape[1]
    n_lines = raw_data['LONGITUDE'].shape[0]
    n_records = eps_file.mdr_counter * n_node_per_line
    idx_nodes = np.arange(eps_file.mdr_counter).repeat(n_node_per_line)

    data = {}
    metadata = {}

    metadata['SPACECRAFT_ID'] = np.int8(eps_file.mphr['SPACECRAFT_ID'][-1])
    metadata['ORBIT_START'] = np.uint32(eps_file.mphr['ORBIT_START'])

    ascat_time = shortcdstime2jd(raw_data['UTC_LINE_NODES'].flatten()['day'],
                                 raw_data['UTC_LINE_NODES'].flatten()['time'])
    data['jd'] = ascat_time[idx_nodes]

    fields = [('SIGMA0_TRIP', long_nan),
              ('INC_ANGLE_TRIP', uint_nan),
              ('AZI_ANGLE_TRIP', int_nan),
              ('KP', uint_nan),
              ('F_LAND', uint_nan)]

    for field in fields:
        data[field[0]] = raw_data[field[0]].reshape(n_records, 3)
        valid = raw_unscaled[field[0]].reshape(n_records, 3) != field[1]
        data[field[0]][~valid] = field[1]

    fields = ['SAT_TRACK_AZI', 'ABS_LINE_NUMBER']
    for field in fields:
        data[field] = raw_data[field].flatten()[idx_nodes]

    fields = [('LONGITUDE', long_nan),
              ('LATITUDE', long_nan),
              ('SOIL_MOISTURE', uint_nan),
              ('SWATH_INDICATOR', byte_nan),
              ('SOIL_MOISTURE_ERROR', uint_nan),
              ('SIGMA40', long_nan),
              ('SIGMA40_ERROR', long_nan),
              ('SLOPE40', long_nan),
              ('SLOPE40_ERROR', long_nan),
              ('DRY_BACKSCATTER', long_nan),
              ('WET_BACKSCATTER', long_nan),
              ('MEAN_SURF_SOIL_MOISTURE', uint_nan),
              ('SOIL_MOISTURE_SENSETIVITY', ulong_nan),
              ('CORRECTION_FLAGS', None),
              ('PROCESSING_FLAGS', None),
              ('AGGREGATED_QUALITY_FLAG', None),
              ('SNOW_COVER_PROBABILITY', None),
              ('FROZEN_SOIL_PROBABILITY', None),
              ('INNUDATION_OR_WETLAND', None),
              ('TOPOGRAPHICAL_COMPLEXITY', None)]

    for field in fields:
        data[field[0]] = raw_data[field[0]].flatten()
        if field[1] is not None:
            valid = raw_unscaled[field[0]].flatten() != field[1]
            data[field[0]][~valid] = field[1]

    # sat_track_azi (uint)
    data['AS_DES_PASS'] = \
        np.array(raw_data['SAT_TRACK_AZI'].flatten()[idx_nodes] < 270)

    # modify longitudes from [0,360] to [-180,180]
    mask = np.logical_and(data['LONGITUDE'] != long_nan,
                          data['LONGITUDE'] > 180)
    data['LONGITUDE'][mask] += -360.

    # modify azimuth from (-180, 180) to (0, 360)
    mask = (data['AZI_ANGLE_TRIP'] != int_nan) & (data['AZI_ANGLE_TRIP'] < 0)
    data['AZI_ANGLE_TRIP'][mask] += 360

    fields = ['PARAM_DB_VERSION', 'WARP_NRT_VERSION']
    for field in fields:
        data[field] = raw_data['PARAM_DB_VERSION'].flatten()[idx_nodes]

    metadata['SPACECRAFT_ID'] = int(eps_file.mphr['SPACECRAFT_ID'][2])

    data['NODE_NUM'] = np.tile((np.arange(n_node_per_line) + 1), n_lines)

    data['LINE_NUM'] = idx_nodes

    return data, metadata


def shortcdstime2jd(days, milliseconds):
    """
    Convert cds time to julian date
    """
    offset = days + (milliseconds / 1000.) / (24. * 60. * 60.)
    return julian_epoch + offset



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""
Readers for SZF data in h5 format.
"""
from __future__ import division

import numpy as np
import h5py

from pygeobase.io_base import ImageBase
from pygeobase.object_base import Image

# 1.1.2000 00:00:00 in jd
julian_epoch = 2451544.5


class AscatL1H5File(ImageBase):
    """
    Read ASCAT L2 SSM File in netCDF format, as downloaded from EUMETSAT

    Parameters
    ----------
    filename : str
        Filename path.
    mode : str, optional
        Opening mode. Default: r
    nc_variables: list, optional
        list of variables to read from netCDF.
        Default: read all available variables
    """

    def __init__(self, filename, mode='r', h5_keys=None, **kwargs):
        """
        Initialization of i/o object.

        """
        super(AscatL1H5File, self).__init__(filename, mode=mode,
                                            **kwargs)
        self.h5_keys = h5_keys
        self.ds = None

    def read(self, timestamp=None):
        """
        reads from the netCDF file given by the filename

        Returns
        -------
        data : pygeobase.object_base.Image
        """

        if self.ds is None:
            self.ds = h5py.File(self.filename)
            while len(self.ds.keys()) == 1:
                self.ds = self.ds[list(self.ds.keys())[0]]
        raw_data = self.ds['DATA']
        raw_metadata = self.ds['METADATA']

        # store data in dictionary
        data = {}
        metadata = {}

        if self.h5_keys is None:
            var_to_read = list(
                raw_data['MDR_1B_FULL_ASCA_Level_1_ARRAY_000001'].dtype.names)
        else:
            var_to_read = self.h5_keys

        # make sure that essential variables are read always:
        if 'LATITUDE_FULL' not in var_to_read:
            var_to_read.append('LATITUDE_FULL')
        if 'LONGITUDE_FULL' not in var_to_read:
            var_to_read.append('LONGITUDE_FULL')

        num_cells = raw_data['MDR_1B_FULL_ASCA_Level_1_ARRAY_000001'][
            'LATITUDE_FULL'].shape[1]
        num_lines = raw_data['MDR_1B_FULL_ASCA_Level_1_ARRAY_000001'][
            'LATITUDE_FULL'].shape[0]

        # read the requested variables and scale them if they have a
        # scaling factor
        # encode() is needed for py3 comparison between str and byte
        for name in var_to_read:
            variable = raw_data['MDR_1B_FULL_ASCA_Level_1_ARRAY_000001'][name]
            if name.encode() in \
                    raw_data['MDR_1B_FULL_ASCA_Level_1_DESCR'].value[
                        'EntryName']:
                var_index = np.where(
                    raw_data['MDR_1B_FULL_ASCA_Level_1_DESCR'][
                        'EntryName'] == name.encode())[0][0]
                if raw_data['MDR_1B_FULL_ASCA_Level_1_DESCR'].value[
                        'Scale Factor'][var_index] != "n/a".encode():
                    sf = 10 ** float(
                        raw_data['MDR_1B_FULL_ASCA_Level_1_DESCR'].value[
                            'Scale Factor'][var_index])
                    variable = variable / sf
            data[name] = variable[:].flatten()
            if len(variable.shape) == 1:
                # If the data is 1D then we repeat it for each cell
                data[name] = np.repeat(data[name], num_cells)

        data['AS_DES_PASS'] = (data['SAT_TRACK_AZI'] < 270).astype(np.uint8)

        for name in raw_metadata.keys():
            for subname in raw_metadata[name].keys():
                if name not in metadata:
                    metadata[name] = dict()
                metadata[name][subname] = raw_metadata[name][subname].value

        # modify longitudes from [0,360] to [-180,180]
        mask = data['LONGITUDE_FULL'] > 180
        data['LONGITUDE_FULL'][mask] += -360.

        if 'AZI_ANGLE_FULL' in var_to_read:
            mask = data['AZI_ANGLE_FULL'] < 0
            data['AZI_ANGLE_FULL'][mask] += 360

        if 'UTC_LOCALISATION-days' in var_to_read and \
                'UTC_LOCALISATION-milliseconds' in var_to_read:
            data['jd'] = shortcdstime2jd(data['UTC_LOCALISATION-days'],
                                         data['UTC_LOCALISATION-milliseconds'])

        set_flags(data)

        fields = ['SPACECRAFT_ID', 'ORBIT_START',
                  'PROCESSOR_MAJOR_VERSION', 'PROCESSOR_MINOR_VERSION',
                  'FORMAT_MAJOR_VERSION', 'FORMAT_MINOR_VERSION'
                  ]
        for field in fields:
            var_index = np.where(
                np.core.defchararray.startswith(
                    metadata['MPHR']['MPHR_TABLE']['EntryName'],
                    field.encode()))[0][0]
            var = metadata['MPHR']['MPHR_TABLE']['EntryValue'][
                var_index].decode()
            if field == 'SPACECRAFT_ID':
                var = var[-1]
            metadata[field] = int(var)

        image_dict = {'img1': {}, 'img2': {}, 'img3': {}, 'img4': {},
                      'img5': {}, 'img6': {}}
        data_full = {'d1': {}, 'd2': {}, 'd3': {}, 'd4': {}, 'd5': {},
                     'd6': {}}

        # separate data into single beam images
        for i in range(1, 7):
            dataset = 'd' + str(i)
            img = 'img' + str(i)
            mask = ((data['BEAM_NUMBER']) == i)
            for field in data:
                data_full[dataset][field] = data[field][mask]

            lon = data_full[dataset].pop('LONGITUDE_FULL')
            lat = data_full[dataset].pop('LATITUDE_FULL')
            image_dict[img] = Image(lon, lat, data_full[dataset], metadata,
                                    timestamp, timekey='jd')

        return image_dict

    def read_masked_data(self, **kwargs):
        """
        It does not make sense to read a orbit file unmasked
        so we only have a masked implementation.
        """
        return self.read(**kwargs)

    def write(self, data):
        raise NotImplementedError()

    def flush(self):
        pass

    def close(self):
        pass


def set_flags(data):
    """
    Compute summary flag for each measurement with a value of 0, 1 or 2
    indicating nominal, slightly degraded or severely degraded data.

    Parameters
    ----------
    data : numpy.ndarray
        SZF data.
    """

    # category:status = 'red': 2, 'amber': 1, 'warning': 0
    flag_status_bit = {'FLAGFIELD_RF1': {'2': [2, 4],
                                         '1': [0, 1, 3]},

                       'FLAGFIELD_RF2': {'2': [0, 1]},

                       'FLAGFIELD_PL': {'2': [0, 1, 2, 3],
                                        '0': [4]},

                       'FLAGFIELD_GEN1': {'2': [1],
                                          '0': [0]},

                       'FLAGFIELD_GEN2': {'2': [2],
                                          '1': [0],
                                          '0': [1]}
                       }

    for flagfield in flag_status_bit.keys():
        # get flag data in binary format to get flags
        unpacked_bits = np.unpackbits(data[flagfield])

        # find indizes where a flag is set
        set_bits = np.where(unpacked_bits == 1)[0]
        if set_bits.size != 0:
            pos_8 = 7 - (set_bits % 8)

            for category in sorted(flag_status_bit[flagfield].keys()):
                if (int(category) == 0) and (flagfield != 'FLAGFIELD_GEN2'):
                    continue

                for bit2check in flag_status_bit[flagfield][category]:
                    pos = np.where(pos_8 == bit2check)[0]
                    data['F_USABLE'] = np.zeros(data['FLAGFIELD_GEN2'].size)
                    data['F_USABLE'][set_bits[pos] // 8] = int(category)

                    # land points
                    if (flagfield == 'FLAGFIELD_GEN2') and (bit2check == 1):
                        data['F_LAND'] = np.zeros(data['FLAGFIELD_GEN2'].size)
                        data['F_LAND'][set_bits[pos] // 8] = 1


def shortcdstime2jd(days, milliseconds):
    """
    Convert cds time to julian date
    """
    offset = days + (milliseconds / 1000.) / (24. * 60. * 60.)
    return julian_epoch + offset



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""
Readers for lvl1b and lvl2 data in nc format.
"""

import os
from datetime import datetime, timedelta

import netCDF4
import numpy as np
import pandas as pd

from pygeobase.io_base import ImageBase
from pygeobase.io_base import MultiTemporalImageBase
from pygeobase.object_base import Image

ref_dt = np.datetime64('1970-01-01')
ref_jd = 2440587.5  # julian date on 1970-01-01 00:00:00


class AscatL1NcFile(ImageBase):
    """
    Read ASCAT L1 File in netCDF format, as downloaded from EUMETSAT

    Parameters
    ----------
    filename : str
        Filename path.
    mode : str, optional
        Opening mode. Default: r
    nc_variables: list, optional
        list of variables to read from netCDF.
        Default: read all available variables
    """

    def __init__(self, filename, mode='r', nc_variables=None, **kwargs):
        """
        Initialization of i/o object.

        """
        super(AscatL1NcFile, self).__init__(filename, mode=mode,
                                            **kwargs)
        self.nc_variables = nc_variables
        self.ds = None

    def read(self, timestamp=None):
        """
        reads from the netCDF file given by the filename

        Returns
        -------
        data : pygeobase.object_base.Image
        """

        if self.ds is None:
            self.ds = netCDF4.Dataset(self.filename)

        if self.nc_variables is None:
            var_to_read = self.ds.variables.keys()
        else:
            var_to_read = self.nc_variables

        # make sure that essential variables are read always:
        if 'latitude' not in var_to_read:
            var_to_read.append('latitude')
        if 'longitude' not in var_to_read:
            var_to_read.append('longitude')

        # store data in dictionary
        dd = {}
        metadata = {}
        beams = ['f_', 'm_', 'a_']

        metadata['sat_id'] = self.ds.platform[-1]
        metadata['orbit_start'] = self.ds.start_orbit_number
        metadata['processor_major_version'] = self.ds.processor_major_version
        metadata['product_minor_version'] = self.ds.product_minor_version
        metadata['format_major_version'] = self.ds.format_major_version
        metadata['format_minor_version'] = self.ds.format_minor_version

        num_cells = self.ds.dimensions['numCells'].size
        for name in var_to_read:
            variable = self.ds.variables[name]

            if len(variable.shape) == 1:
                # If the data is 1D then we repeat it for each cell
                dd[name] = variable[:].flatten()
                dd[name] = np.repeat(dd[name], num_cells)
            elif len(variable.shape) == 2:
                dd[name] = variable[:].flatten()
            elif len(variable.shape) == 3:
                # length of 3 means it is triplet data, so we split it
                for i, beam in enumerate(beams):
                    dd[beam + name] = variable[:, :, i].flatten()
                    if name == 'azi_angle_trip':
                        mask = dd[beam + name] < 0
                        dd[beam + name][mask] += 360
            else:
                raise RuntimeError("Unexpected variable shape.")

            if name == 'utc_line_nodes':
                utc_dates = netCDF4.num2date(
                    dd[name], variable.units).astype('datetime64[ns]')
                dd['jd'] = (utc_dates - ref_dt)/np.timedelta64(1, 'D') + ref_jd

        dd['as_des_pass'] = (dd['sat_track_azi'] < 270).astype(np.uint8)

        longitude = dd.pop('longitude')
        latitude = dd.pop('latitude')

        n_records = latitude.shape[0]
        n_lines = n_records // num_cells
        dd['node_num'] = np.tile((np.arange(num_cells) + 1), n_lines)
        dd['line_num'] = np.arange(n_lines).repeat(num_cells)

        return Image(longitude, latitude, dd, metadata, timestamp,
                     timekey='utc_line_nodes')

    def write(self, data):
        raise NotImplementedError()

    def flush(self):
        pass

    def close(self):
        pass


class AscatL2SsmNcFile(ImageBase):
    """
    Read ASCAT L2 SSM File in netCDF format, as downloaded from EUMETSAT

    Parameters
    ----------
    filename : str
        Filename path.
    mode : str, optional
        Opening mode. Default: r
    nc_variables: list, optional
        list of variables to read from netCDF.
        Default: read all available variables
    """

    def __init__(self, filename, mode='r', nc_variables=None, **kwargs):
        """
        Initialization of i/o object.

        """
        super(AscatL2SsmNcFile, self).__init__(filename, mode=mode,
                                               **kwargs)
        self.nc_variables = nc_variables
        self.ds = None

    def read(self, timestamp=None, ssm_masked=False):
        """
        reads from the netCDF file given by the filename

        Returns
        -------
        data : pygeobase.object_base.Image
        """

        if self.ds is None:
            self.ds = netCDF4.Dataset(self.filename)

        if self.nc_variables is None:
            var_to_read = self.ds.variables.keys()
        else:
            var_to_read = self.nc_variables

        # make sure that essential variables are read always:
        if 'latitude' not in var_to_read:
            var_to_read.append('latitude')
        if 'longitude' not in var_to_read:
            var_to_read.append('longitude')

        # store data in dictionary
        dd = {}
        metadata = {}

        metadata['sat_id'] = self.ds.platform_long_name[-1]
        metadata['orbit_start'] = self.ds.start_orbit_number
        metadata['processor_major_version'] = self.ds.processor_major_version
        metadata['product_minor_version'] = self.ds.product_minor_version
        metadata['format_major_version'] = self.ds.format_major_version
        metadata['format_minor_version'] = self.ds.format_minor_version

        num_cells = self.ds.dimensions['numCells'].size
        for name in var_to_read:
            variable = self.ds.variables[name]
            dd[name] = variable[:].flatten()
            if len(variable.shape) == 1:
                # If the data is 1D then we repeat it for each cell
                dd[name] = np.repeat(dd[name], num_cells)

            if name == 'utc_line_nodes':
                utc_dates = netCDF4.num2date(
                    dd[name], variable.units).astype('datetime64[ns]')
                dd['jd'] = (utc_dates - ref_dt)/np.timedelta64(1, 'D') + ref_jd

        # if the ssm_masked is True we mask out data with missing ssm value
        if 'soil_moisture' in dd and ssm_masked is True:
            # mask all the arrays based on fill_value of latitude
            valid_data = ~dd['soil_moisture'].mask
            for name in dd:
                dd[name] = dd[name][valid_data]

        longitude = dd.pop('longitude')
        latitude = dd.pop('latitude')

        n_records = latitude.shape[0]
        n_lines = n_records // num_cells
        dd['node_num'] = np.tile((np.arange(num_cells) + 1), n_lines)
        dd['line_num'] = np.arange(n_lines).repeat(num_cells)

        dd['as_des_pass'] = (dd['sat_track_azi'] < 270).astype(np.uint8)

        return Image(longitude, latitude, dd, metadata, timestamp,
                     timekey='utc_line_nodes')

    def write(self, data):
        raise NotImplementedError()

    def flush(self):
        pass

    def close(self):
        pass


class AscatL2SsmNc(MultiTemporalImageBase):
    """
    Class for reading HSAF ASCAT SSM images in netCDF format.
    The images have to be uncompressed in the following folder structure

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        if the files are stored in folders by month as is the standard on
        the HSAF FTP Server then please specify the string that should be used
        in datetime.datetime.strftime Default: ''
    day_search_str: string, optional
        to provide an iterator over all images of a day the method
        _get_possible_timestamps looks for all available images on a day on the
        harddisk. This string is used in datetime.datetime.strftime and in
        glob.glob to search for all files on a day.
    file_search_str: string, optional
        this string is used in datetime.datetime.strftime and glob.glob to find
        a 3 minute bufr file by the exact date.
    datetime_format: string, optional
        datetime format by which {datetime} will be replaced in file_search_str
    nc_variables: list, optional
        list of variables to read from netCDF.
        Default: read all available variables
    """

    def __init__(self, path, month_path_str='',
                 day_search_str='W_XX-EUMETSAT-Darmstadt,'
                                'SURFACE+SATELLITE,METOPA+'
                                'ASCAT_C_EUMP_%Y%m%d*_125_ssm_l2.nc',
                 file_search_str='W_XX-EUMETSAT-Darmstadt,'
                                 'SURFACE+SATELLITE,METOPA+'
                                 'ASCAT_C_EUMP_{datetime}*_125_ssm_l2.nc',
                 datetime_format='%Y%m%d%H%M%S',
                 filename_datetime_format=(62, 76, '%Y%m%d%H%M%S'),
                 nc_variables=None):
        self.path = path
        self.month_path_str = month_path_str
        self.day_search_str = day_search_str
        self.file_search_str = file_search_str
        self.filename_datetime_format = filename_datetime_format
        super(AscatL2SsmNc, self).__init__(path, AscatL2SsmNcFile,
                                           subpath_templ=[month_path_str],
                                           fname_templ=file_search_str,
                                           datetime_format=datetime_format,
                                           exact_templ=False,
                                           ioclass_kws={
                                               'nc_variables': nc_variables})

    def _get_orbit_start_date(self, filename):
        orbit_start_str = \
            os.path.basename(filename)[self.filename_datetime_format[0]:
                                       self.filename_datetime_format[1]]
        return datetime.strptime(orbit_start_str,
                                 self.filename_datetime_format[2])

    def tstamps_for_daterange(self, startdate, enddate):
        """
        Get the timestamps as datetime array that are possible for the
        given day, if the timestamps are

        For this product it is not fixed but has to be looked up from
        the hard disk since bufr files are not regular spaced and only
        europe is in this product. For a global product a 3 minute
        spacing could be used as a fist approximation

        Parameters
        ----------
        startdate : datetime.date or datetime.datetime
            start date
        enddate : datetime.date or datetime.datetime
            end date

        Returns
        -------
        dates : list
            list of datetimes
        """
        file_list = []
        delta_all = enddate - startdate
        timestamps = []

        for i in range(delta_all.days + 1):
            timestamp = startdate + timedelta(days=i)

            files = self._search_files(
                timestamp, custom_templ=self.day_search_str)

            file_list.extend(sorted(files))

        for filename in file_list:
            timestamps.append(self._get_orbit_start_date(filename))

        timestamps = [dt for dt in timestamps if startdate <= dt <= enddate]
        return timestamps



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import unittest
from datetime import datetime

import numpy as np

from ascat.tuw import AscatVodTs, Ascat_SSM


class TestAscat(unittest.TestCase):

    def setUp(self):
        self.ascat_folder = os.path.join(os.path.dirname(__file__),
                                         'ascat_test_data', 'tuw', 'ascat', 'ssm')

        self.ascat_adv_folder = os.path.join(os.path.dirname(__file__),
                                             'ascat_test_data', 'tuw',
                                             'advisory_flags')

        self.ascat_grid_folder = os.path.join(os.path.dirname(__file__),
                                              'ascat_test_data', 'tuw',
                                              'grid')

        # init the ASCAT_SSM reader with the paths
        self.ascat_SSM_reader = Ascat_SSM(
            self.ascat_folder, self.ascat_grid_folder,
            advisory_flags_path=self.ascat_adv_folder)

    def test_read_ssm(self):

        gpi = 2329253
        result = self.ascat_SSM_reader.read_ssm(gpi)
        assert result.gpi == gpi
        assert result.longitude == 14.28413
        assert result.latitude == 45.698074
        assert sorted(list(result.data.columns)) == sorted([
            'ERR', 'SSF', 'SSM', 'frozen_prob', 'snow_prob'])
        assert len(result.data) == 2058
        assert result.data.iloc[15].name == datetime(2007, 1, 15, 19)
        assert result.data.iloc[15]['ERR'] == 7
        assert result.data.iloc[15]['SSF'] == 1
        assert result.data.iloc[15]['SSM'] == 53
        assert result.data.iloc[15]['frozen_prob'] == 29
        assert result.data.iloc[15]['snow_prob'] == 0

    def test_neighbor_search(self):

        self.ascat_SSM_reader._load_grid_info()
        gpi, distance = self.ascat_SSM_reader.grid.find_nearest_gpi(
            3.25, 46.13)
        assert gpi == 2346869
        np.testing.assert_approx_equal(distance, 2267.42, significant=2)


class TestAscatVodTs(unittest.TestCase):

    def setUp(self):
        self.ascat_folder = os.path.join(os.path.dirname(__file__),
                                         'ascat_test_data', 'tuw', 'ascat', 'vod')

        self.ascat_grid_folder = os.path.join(os.path.dirname(__file__),
                                              'ascat_test_data', 'hsaf', 'grid')

        # init the ASCAT_SSM reader with the paths
        self.ascat_VOD_reader = AscatVodTs(self.ascat_folder,
                                           self.ascat_grid_folder)

    def test_read_vod(self):
        gpi = 2199945
        data = self.ascat_VOD_reader.read(gpi)
        lon, lat = self.ascat_VOD_reader.grid.gpi2lonlat(gpi)
        np.testing.assert_approx_equal(lon, 9.1312, significant=4)
        np.testing.assert_approx_equal(lat, 42.5481, significant=4)

        assert list(data.columns) == ['vod']
        assert len(data) == 4018
        assert data.iloc[15].name == datetime(2007, 1, 16, 12, 0, 0)
        assert data.iloc[15]['vod'] == np.float32(0.62470651)

    def test_neighbor_search(self):
        gpi, distance = self.ascat_VOD_reader.grid.find_nearest_gpi(
            3.25, 46.13)
        assert gpi == 2346869
        np.testing.assert_approx_equal(distance, 2267.42, significant=2)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np


def db2lin(val):
    """
    Converting from linear to dB domain.

    Parameters
    ----------
    val : numpy.ndarray
        Values in dB domain.

    Returns
    -------
    val : numpy.ndarray
        Values in linear domain.
    """
    return 10 ** (val / 10.)


def lin2db(val):
    """
    Converting from linear to dB domain.

    Parameters
    ----------
    val : numpy.ndarray
        Values in linear domain.

    Returns
    -------
    val : numpy.ndarray
        Values in dB domain.
    """
    return 10. * np.log10(val)


def get_window_radius(window, hp_radius):
    """
    Calculates the required radius of a window function in order to achieve
    the provided half power radius.

    Parameters
    ----------
    window : string
        Window function name.
        Current supported windows:
            - Hamming
            - Boxcar
    hp_radius : float32
        Half power radius. Radius of window function for weight
        equal to 0.5 (-3 dB). In the spatial domain this corresponds to
        half of the spatial resolution one would like to achieve with the
        given window.
    Returns
    -------
    r : float32
        Window radius needed to achieve the given half power radius

    """
    window = window.lower()
    hp_weight = 0.5
    if window == 'hamming':
        alpha = 0.54
        r = (np.pi * hp_radius) / np.arccos((hp_weight-alpha) / (1-alpha))
    elif window == 'boxcar':
        r = hp_radius
    else:
        raise ValueError('Window name not supported.')

    return r


def hamming_window(radius, distances):
    """
    Hamming window filter.

    Parameters
    ----------
    radius : float32
        Radius of the window.
    distances : numpy.ndarray
        Array with distances.

    Returns
    -------
    weights : numpy.ndarray
        Distance weights.
    tw : float32
        Sum of weigths.
    """
    alpha = 0.54
    weights = alpha + (1 - alpha) * np.cos(np.pi / radius * distances)

    return weights, np.sum(weights)


def boxcar(radius, distance):
    """
    Boxcar filter

    Parameters
    ----------
    n : int
        Length.

    Returns
    -------
    weights : numpy.ndarray
        Distance weights.
    tw : float32
        Sum of weigths.
    """
    weights = np.zeros(distance.size)
    weights[distance <= radius] = 1.

    return weights, np.sum(weights)


def get_window_weights(window, radius, distance, norm=False):
    """
    Function returning weights for the provided window function

    Parameters
    ----------
    window : str
        Window function name
    radius : float
        Radius of the window.
    distance : numpy.ndarray
        Distance array
    norm : boolean
        If true, normalised weights will be returned.

    Returns
    -------
    weights : numpy.ndarray
        Weights according to distances and given window function

    """
    if window == 'hamming':
        weights, w_sum = hamming_window(radius, distance)
    elif window == 'boxcar':
        weights, w_sum = boxcar(radius, distance)
    else:
        raise ValueError('Window name not supported.')

    if norm is True:
        weights = weights / w_sum

    return weights



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Test data downloaded from EUMETSAT UMARF
"""

import unittest
import datetime
import numpy as np
import numpy.testing as nptest
import os
import sys
import pytest

from ascat.eumetsat import AscatL2Ssm125
from ascat.eumetsat import AscatL2Ssm125PDU
from ascat.eumetsat import AscatL2Ssm250
from ascat.eumetsat import AscatL2Ssm250PDU
from ascat.eumetsat import AscatL2Ssm125Nc
from ascat.eumetsat import AscatL2Ssm250Nc


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_ASCAT_A_L2_SSM_125_BUFR(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(os.path.dirname(__file__),  'ascat_test_data',
                                 'eumetsat', 'ASCAT_L2_SM_125', 'bufr')
        self.reader_A = AscatL2Ssm125(data_path, 'Metop_A')
        self.reader_B = AscatL2Ssm125(data_path, 'Metop_B')

    def tearDown(self):
        self.reader_A = None
        self.reader_B = None

    def test_offset_getting(self):
        """
        test getting the image offsets for a known day
        """
        timestamps = self.reader_A.tstamps_for_daterange(
            datetime.datetime(2017, 2, 20), datetime.datetime(2017, 2, 21))
        timestamps_should = [datetime.datetime(2017, 2, 20, 4, 15),
                             datetime.datetime(2017, 2, 20, 5, 57)]
        assert sorted(timestamps) == sorted(timestamps_should)

        timestamps = self.reader_B.tstamps_for_daterange(
            datetime.datetime(2017, 2, 20), datetime.datetime(2017, 2, 21))
        timestamps_should = [datetime.datetime(2017, 2, 20, 5, 9)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader_A.read(
            datetime.datetime(2017, 2, 20, 4, 15))

        ssm_should = np.array([3., 0., 0., 0., 0., 0., 0., 0., 0., 1.8, 3.3,
                               4.8, 4.3, 2.5, 0., 3.8, 5.8, 1.5, 2.4, 4.1, 2.3,
                               2.7, 5.6, 5.5, 4.9])

        lats_should = np.array([62.60224, 62.67133, 62.74015, 62.80871, 62.877,
                                62.94502, 63.01276, 63.08024, 63.14743,
                                63.21435, 63.28098, 63.34734, 63.41341,
                                63.47919, 63.54468, 63.60988, 63.67479,
                                63.7394, 63.80372, 63.86773, 63.93144,
                                63.99485, 64.05795, 64.12075, 64.18323])

        ssm_mean_should = np.array([21.3, 21.3, 21.4, 22.4, 23.4, 24.5, 26.,
                                    27.1, 27., 26.6, 27.1, 27.6, 27.4, 26.7,
                                    26.5, 27.5, 28.2, 28.4, 28.8, 29.2, 30.,
                                    31., 31.3, 31.9, 32.1])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               :25], ssm_should, atol=0.01)
        nptest.assert_allclose(data['Mean Surface Soil Moisture'][:25],
                               ssm_mean_should,
                               atol=0.01)

        data, meta, timestamp, lons, lats, time_var = self.reader_B.read(
            datetime.datetime(2017, 2, 20, 5, 9))

        ssm_should = np.array([29.2, 30.2, 35.7, 38.6, 37.5, 37.6, 40.5, 44.5,
                               40.7, 39.7, 41.5, 38.8, 34.5, 36.8, 39.4, 41.2,
                               42.4, 42.9, 39.3, 30.5, 26.7, 26.5, 26.7, 23.9,
                               26.2])

        lats_should = np.array([64.74398, 64.81854, 64.89284, 64.96688,
                                65.04066, 65.11416, 65.18739, 65.26036,
                                65.33304, 65.40545, 65.47758, 65.54942,
                                65.62099, 65.69226, 65.76324, 65.83393,
                                65.90432, 65.97442, 66.04422, 66.11371,
                                66.1829, 66.25177, 66.32034, 66.38859,
                                66.45653])

        ssm_mean_should = np.array([36.7, 35.4, 33.4, 32.5, 32.5, 32., 31.2,
                                    29.4, 28.7, 27.6, 25.8, 25.4, 25.5, 25.3,
                                    24.4, 23.4, 22.3, 21.3, 20.4, 20.4, 19.9,
                                    19.7, 20.3, 21.5, 22.9])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               :25], ssm_should, atol=0.01)
        nptest.assert_allclose(data['Mean Surface Soil Moisture'][:25],
                               ssm_mean_should,
                               atol=0.01)


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_ASCAT_L2_SSM_125_BUFR_PDU(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(os.path.dirname(__file__),  'ascat_test_data',
                                 'eumetsat', 'ASCAT_L2_SM_125', 'PDU')
        self.reader_A = AscatL2Ssm125PDU(data_path, 'Metop_A')
        self.reader_B = AscatL2Ssm125PDU(data_path, 'Metop_B')

    def tearDown(self):
        self.reader_A = None
        self.reader_B = None

    def test_offset_getting(self):
        """
        test getting the image offsets for a known day
        """
        timestamps = self.reader_A.tstamps_for_daterange(
            datetime.datetime(2017, 2, 20), datetime.datetime(2017, 2, 21))
        timestamps_should = [datetime.datetime(2017, 2, 20, 4, 15),
                             datetime.datetime(2017, 2, 20, 4, 18),
                             datetime.datetime(2017, 2, 20, 4, 21)]
        assert sorted(timestamps) == sorted(timestamps_should)

        timestamps = self.reader_B.tstamps_for_daterange(
            datetime.datetime(2017, 2, 20), datetime.datetime(2017, 2, 21))
        timestamps_should = [datetime.datetime(2017, 2, 20, 5, 9),
                             datetime.datetime(2017, 2, 20, 5, 12),
                             datetime.datetime(2017, 2, 20, 5, 15)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader_A.read(
            datetime.datetime(2017, 2, 20, 4, 15))

        ssm_should = np.array([3., 0., 0., 0., 0., 0., 0., 0., 0., 1.8, 3.3,
                               4.8, 4.3, 2.5, 0., 3.8, 5.8, 1.5, 2.4, 4.1, 2.3,
                               2.7, 5.6, 5.5, 4.9])

        lats_should = np.array([62.60224, 62.67133, 62.74015, 62.80871, 62.877,
                                62.94502, 63.01276, 63.08024, 63.14743,
                                63.21435, 63.28098, 63.34734, 63.41341,
                                63.47919, 63.54468, 63.60988, 63.67479,
                                63.7394, 63.80372, 63.86773, 63.93144,
                                63.99485, 64.05795, 64.12075, 64.18323])

        ssm_mean_should = np.array([21.3, 21.3, 21.4, 22.4, 23.4, 24.5, 26.,
                                    27.1, 27., 26.6, 27.1, 27.6, 27.4, 26.7,
                                    26.5, 27.5, 28.2, 28.4, 28.8, 29.2, 30.,
                                    31., 31.3, 31.9, 32.1])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               :25], ssm_should, atol=0.01)
        nptest.assert_allclose(data['Mean Surface Soil Moisture'][:25],
                               ssm_mean_should,
                               atol=0.01)

        data, meta, timestamp, lons, lats, time_var = self.reader_B.read(
            datetime.datetime(2017, 2, 20, 5, 9))

        ssm_should = np.array([29.2, 30.2, 35.7, 38.6, 37.5, 37.6, 40.5, 44.5,
                               40.7, 39.7, 41.5, 38.8, 34.5, 36.8, 39.4, 41.2,
                               42.4, 42.9, 39.3, 30.5, 26.7, 26.5, 26.7, 23.9,
                               26.2])

        lats_should = np.array([64.74398, 64.81854, 64.89284, 64.96688,
                                65.04066, 65.11416, 65.18739, 65.26036,
                                65.33304, 65.40545, 65.47758, 65.54942,
                                65.62099, 65.69226, 65.76324, 65.83393,
                                65.90432, 65.97442, 66.04422, 66.11371,
                                66.1829, 66.25177, 66.32034, 66.38859,
                                66.45653])

        ssm_mean_should = np.array([36.7, 35.4, 33.4, 32.5, 32.5, 32., 31.2,
                                    29.4, 28.7, 27.6, 25.8, 25.4, 25.5, 25.3,
                                    24.4, 23.4, 22.3, 21.3, 20.4, 20.4, 19.9,
                                    19.7, 20.3, 21.5, 22.9])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               :25], ssm_should, atol=0.01)
        nptest.assert_allclose(data['Mean Surface Soil Moisture'][:25],
                               ssm_mean_should,
                               atol=0.01)


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_ASCAT_L2_SSM_250_BUFR(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(os.path.dirname(__file__),  'ascat_test_data',
                                 'eumetsat', 'ASCAT_L2_SM_250', 'bufr')
        self.reader_A = AscatL2Ssm250(data_path, 'Metop_A')
        self.reader_B = AscatL2Ssm250(data_path, 'Metop_B')

    def tearDown(self):
        self.reader_A = None
        self.reader_B = None

    def test_offset_getting(self):
        """
        test getting the image offsets for a known day
        """
        timestamps = self.reader_A.tstamps_for_daterange(
            datetime.datetime(2017, 2, 20), datetime.datetime(2017, 2, 21))
        timestamps_should = [datetime.datetime(2017, 2, 20, 4, 15),
                             datetime.datetime(2017, 2, 20, 5, 57)]
        assert sorted(timestamps) == sorted(timestamps_should)

        timestamps = self.reader_B.tstamps_for_daterange(
            datetime.datetime(2017, 2, 20), datetime.datetime(2017, 2, 21))
        timestamps_should = [datetime.datetime(2017, 2, 20, 5, 9)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader_A.read(
            datetime.datetime(2017, 2, 20, 4, 15))

        ssm_should = np.array([1.8, 0., 0., 0., 0., 4.6, 2.8, 4., 4.1, 4.2,
                               4.7, 5.4, 7.1, 7.1, 8.2, 9.2, 14.5, 15.4, 14.3,
                               17.7, 25.5, 36.9, 37.8, 39.4, 24.1])

        lats_should = np.array([62.60224, 62.74015, 62.877, 63.01276,
                                63.14743, 63.28098, 63.41341, 63.54468,
                                63.67479, 63.80372, 63.93144, 64.05795,
                                64.18323, 64.30725, 64.42999, 64.55145,
                                64.6716, 64.79042, 64.9079, 65.02401,
                                65.13873, 67.85438, 67.91597, 67.97556,
                                68.03314])

        ssm_mean_should = np.array([21.3, 21.4, 23.4, 26., 27., 27.1, 27.4,
                                    26.5, 28.2, 28.8, 30., 31.3, 32.1, 30.6,
                                    27.8, 28.9, 29.5, 32.1, 33.8, 32.9, 28.9,
                                    41.1, 40.8, 34.4, 31.])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               :25], ssm_should, atol=0.01)
        nptest.assert_allclose(data['Mean Surface Soil Moisture'][:25],
                               ssm_mean_should,
                               atol=0.01)

        data, meta, timestamp, lons, lats, time_var = self.reader_B.read(
            datetime.datetime(2017, 2, 20, 5, 9))

        ssm_should = np.array([28.8, 31., 35.8, 38.7, 39.3, 38.9, 39.6, 40.7,
                               40.9, 35.5, 28.7, 25.2, 25.8, 27.2, 26.3, 29.1,
                               30., 27.1, 25.5, 23.9, 25.7, 44.9, 38.7, 36.7,
                               40.6])

        lats_should = np.array([64.74398, 64.89284, 65.04066, 65.18739,
                                65.33304, 65.47758, 65.62099, 65.76324,
                                65.90432, 66.04422, 66.1829, 66.32034,
                                66.45653, 66.59144, 66.72505, 66.85734,
                                66.98829, 67.11787, 67.24605, 67.37283,
                                67.49816, 70.48423, 70.55154, 70.61658,
                                70.67934])

        ssm_mean_should = np.array([36.7, 33.4, 32.5, 31.2, 28.7, 25.8, 25.5,
                                    24.4, 22.3, 20.4, 19.9, 20.3, 22.9, 23.7,
                                    23.5, 22.2, 22.2, 22.4, 25.3, 27.8, 27.7,
                                    30.7, 30.7, 31.6, 33.6])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               :25], ssm_should, atol=0.01)
        nptest.assert_allclose(data['Mean Surface Soil Moisture'][:25],
                               ssm_mean_should,
                               atol=0.01)


pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")


class Test_ASCAT_L2_SSM_250_BUFR_PDU(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(os.path.dirname(__file__),  'ascat_test_data',
                                 'eumetsat', 'ASCAT_L2_SM_250', 'PDU')
        self.reader_A = AscatL2Ssm250PDU(data_path, 'Metop_A')
        self.reader_B = AscatL2Ssm250PDU(data_path, 'Metop_B')

    def tearDown(self):
        self.reader_A = None
        self.reader_B = None

    def test_offset_getting(self):
        """
        test getting the image offsets for a known day
        """
        timestamps = self.reader_A.tstamps_for_daterange(
            datetime.datetime(2017, 2, 20), datetime.datetime(2017, 2, 21))
        timestamps_should = [datetime.datetime(2017, 2, 20, 4, 15),
                             datetime.datetime(2017, 2, 20, 4, 18),
                             datetime.datetime(2017, 2, 20, 4, 21)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader_A.read(
            datetime.datetime(2017, 2, 20, 4, 15))

        ssm_should = np.array([1.8, 0., 0., 0., 0., 4.6, 2.8, 4., 4.1, 4.2,
                               4.7, 5.4, 7.1, 7.1, 8.2, 9.2, 14.5, 15.4, 14.3,
                               17.7, 25.5, 36.9, 37.8, 39.4, 24.1])

        lats_should = np.array([62.60224, 62.74015, 62.877, 63.01276,
                                63.14743, 63.28098, 63.41341, 63.54468,
                                63.67479, 63.80372, 63.93144, 64.05795,
                                64.18323, 64.30725, 64.42999, 64.55145,
                                64.6716, 64.79042, 64.9079, 65.02401,
                                65.13873, 67.85438, 67.91597, 67.97556,
                                68.03314])

        ssm_mean_should = np.array([21.3, 21.4, 23.4, 26., 27., 27.1, 27.4,
                                    26.5, 28.2, 28.8, 30., 31.3, 32.1, 30.6,
                                    27.8, 28.9, 29.5, 32.1, 33.8, 32.9, 28.9,
                                    41.1, 40.8, 34.4, 31.])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               :25], ssm_should, atol=0.01)
        nptest.assert_allclose(data['Mean Surface Soil Moisture'][:25],
                               ssm_mean_should,
                               atol=0.01)

        data, meta, timestamp, lons, lats, time_var = self.reader_B.read(
            datetime.datetime(2017, 2, 20, 5, 9))

        ssm_should = np.array([28.8, 31., 35.8, 38.7, 39.3, 38.9, 39.6, 40.7,
                               40.9, 35.5, 28.7, 25.2, 25.8, 27.2, 26.3, 29.1,
                               30., 27.1, 25.5, 23.9, 25.7, 44.9, 38.7, 36.7,
                               40.6])

        lats_should = np.array([64.74398, 64.89284, 65.04066, 65.18739,
                                65.33304, 65.47758, 65.62099, 65.76324,
                                65.90432, 66.04422, 66.1829, 66.32034,
                                66.45653, 66.59144, 66.72505, 66.85734,
                                66.98829, 67.11787, 67.24605, 67.37283,
                                67.49816, 70.48423, 70.55154, 70.61658,
                                70.67934])

        ssm_mean_should = np.array([36.7, 33.4, 32.5, 31.2, 28.7, 25.8, 25.5,
                                    24.4, 22.3, 20.4, 19.9, 20.3, 22.9, 23.7,
                                    23.5, 22.2, 22.2, 22.4, 25.3, 27.8, 27.7,
                                    30.7, 30.7, 31.6, 33.6])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               :25], ssm_should, atol=0.01)
        nptest.assert_allclose(data['Mean Surface Soil Moisture'][:25],
                               ssm_mean_should,
                               atol=0.01)


class Test_ASCAT_L2_SSM_125_NC(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(os.path.dirname(__file__),  'ascat_test_data',
                                 'eumetsat', 'ASCAT_L2_SM_125', 'nc')
        self.reader_A = AscatL2Ssm125Nc(data_path, 'Metop_A')
        self.reader_B = AscatL2Ssm125Nc(data_path, 'Metop_B')

    def tearDown(self):
        self.reader_A = None
        self.reader_B = None

    def test_offset_getting(self):
        """
        test getting the image offsets for a known day
        """
        timestamps = self.reader_A.tstamps_for_daterange(
            datetime.datetime(2017, 2, 20), datetime.datetime(2017, 2, 21))
        timestamps_should = [datetime.datetime(2017, 2, 20, 4, 15),
                             datetime.datetime(2017, 2, 20, 5, 57)]
        assert sorted(timestamps) == sorted(timestamps_should)

        timestamps = self.reader_B.tstamps_for_daterange(
            datetime.datetime(2017, 2, 20), datetime.datetime(2017, 2, 21))
        timestamps_should = [datetime.datetime(2017, 2, 20, 5, 9)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader_A.read(
            datetime.datetime(2017, 2, 20, 4, 15))

        ssm_should = np.array([3., 0., 0., 0., 0., 0., 0., 0., 0., 1.8, 3.3,
                               4.8, 4.3, 2.5, 0., 3.8, 5.8, 1.5, 2.4, 4.1, 2.3,
                               2.7, 5.6, 5.5, 4.9])

        lats_should = np.array([62.60224, 62.67133, 62.74015, 62.80871, 62.877,
                                62.94502, 63.01276, 63.08024, 63.14743,
                                63.21435, 63.28098, 63.34734, 63.41341,
                                63.47919, 63.54468, 63.60988, 63.67479,
                                63.7394, 63.80372, 63.86773, 63.93144,
                                63.99485, 64.05795, 64.12075, 64.18323])

        ssm_mean_should = np.array([21.3, 21.3, 21.4, 22.4, 23.4, 24.5, 26.,
                                    27.1, 27., 26.6, 27.1, 27.6, 27.4, 26.7,
                                    26.5, 27.5, 28.2, 28.4, 28.8, 29.2, 30.,
                                    31., 31.3, 31.9, 32.1])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['soil_moisture'][
                               :25], ssm_should, atol=0.1)
        nptest.assert_allclose(data['mean_soil_moisture'][:25],
                               ssm_mean_should,
                               atol=0.1)

        data, meta, timestamp, lons, lats, time_var = self.reader_B.read(
            datetime.datetime(2017, 2, 20, 5, 9))

        ssm_should = np.array([29.2, 30.2, 35.7, 38.6, 37.5, 37.6, 40.5, 44.5,
                               40.7, 39.7, 41.5, 38.8, 34.5, 36.8, 39.4, 41.2,
                               42.4, 42.9, 39.3, 30.5, 26.7, 26.5, 26.7, 23.9,
                               26.2])

        lats_should = np.array([64.74398, 64.81854, 64.89284, 64.96688,
                                65.04066, 65.11416, 65.18739, 65.26036,
                                65.33304, 65.40545, 65.47758, 65.54942,
                                65.62099, 65.69226, 65.76324, 65.83393,
                                65.90432, 65.97442, 66.04422, 66.11371,
                                66.1829, 66.25177, 66.32034, 66.38859,
                                66.45653])

        ssm_mean_should = np.array([36.7, 35.4, 33.4, 32.5, 32.5, 32., 31.2,
                                    29.4, 28.7, 27.6, 25.8, 25.4, 25.5, 25.3,
                                    24.4, 23.4, 22.3, 21.3, 20.4, 20.4, 19.9,
                                    19.7, 20.3, 21.5, 22.9])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['soil_moisture'][
                               :25], ssm_should, atol=0.1)
        nptest.assert_allclose(data['mean_soil_moisture'][:25],
                               ssm_mean_should,
                               atol=0.1)


class Test_ASCAT_L2_SSM_250_NC(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(os.path.dirname(__file__),  'ascat_test_data',
                                 'eumetsat', 'ASCAT_L2_SM_250', 'nc')
        self.reader_A = AscatL2Ssm250Nc(data_path, 'Metop_A')
        self.reader_B = AscatL2Ssm250Nc(data_path, 'Metop_B')

    def tearDown(self):
        self.reader_A = None
        self.reader_B = None

    def test_offset_getting(self):
        """
        test getting the image offsets for a known day
        """
        timestamps = self.reader_A.tstamps_for_daterange(
            datetime.datetime(2017, 2, 20), datetime.datetime(2017, 2, 21))
        timestamps_should = [datetime.datetime(2017, 2, 20, 4, 15),
                             datetime.datetime(2017, 2, 20, 5, 57)]
        assert sorted(timestamps) == sorted(timestamps_should)

        timestamps = self.reader_B.tstamps_for_daterange(
            datetime.datetime(2017, 2, 20), datetime.datetime(2017, 2, 21))
        timestamps_should = [datetime.datetime(2017, 2, 20, 5, 9)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader_A.read(
            datetime.datetime(2017, 2, 20, 4, 15))

        ssm_should = np.array([1.8, 0., 0., 0., 0., 4.6, 2.8, 4., 4.1, 4.2,
                               4.7, 5.4, 7.1, 7.1, 8.2, 9.2, 14.5, 15.4, 14.3,
                               17.7, 25.5, 36.9, 37.8, 39.4, 24.1])

        lats_should = np.array([62.60224, 62.74015, 62.877, 63.01276,
                                63.14743, 63.28098, 63.41341, 63.54468,
                                63.67479, 63.80372, 63.93144, 64.05795,
                                64.18323, 64.30725, 64.42999, 64.55145,
                                64.6716, 64.79042, 64.9079, 65.02401,
                                65.13873, 67.85438, 67.91597, 67.97556,
                                68.03314])

        ssm_mean_should = np.array([21.3, 21.4, 23.4, 26., 27., 27.1, 27.4,
                                    26.5, 28.2, 28.8, 30., 31.3, 32.1, 30.6,
                                    27.8, 28.9, 29.5, 32.1, 33.8, 32.9, 28.9,
                                    41.1, 40.8, 34.4, 31.])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['soil_moisture'][
                               :25], ssm_should, atol=0.1)
        nptest.assert_allclose(data['mean_soil_moisture'][:25],
                               ssm_mean_should,
                               atol=0.1)

        data, meta, timestamp, lons, lats, time_var = self.reader_B.read(
            datetime.datetime(2017, 2, 20, 5, 9))

        ssm_should = np.array([28.8, 31., 35.8, 38.7, 39.3, 38.9, 39.6, 40.7,
                               40.9, 35.5, 28.7, 25.2, 25.8, 27.2, 26.3, 29.1,
                               30., 27.1, 25.5, 23.9, 25.7, 44.9, 38.7, 36.7,
                               40.6])

        lats_should = np.array([64.74398, 64.89284, 65.04066, 65.18739,
                                65.33304, 65.47758, 65.62099, 65.76324,
                                65.90432, 66.04422, 66.1829, 66.32034,
                                66.45653, 66.59144, 66.72505, 66.85734,
                                66.98829, 67.11787, 67.24605, 67.37283,
                                67.49816, 70.48423, 70.55154, 70.61658,
                                70.67934])

        ssm_mean_should = np.array([36.7, 33.4, 32.5, 31.2, 28.7, 25.8, 25.5,
                                    24.4, 22.3, 20.4, 19.9, 20.3, 22.9, 23.7,
                                    23.5, 22.2, 22.2, 22.4, 25.3, 27.8, 27.7,
                                    30.7, 30.7, 31.6, 33.6])

        nptest.assert_allclose(lats[:25], lats_should, atol=1e-5)
        nptest.assert_allclose(data['soil_moisture'][
                               :25], ssm_should, atol=0.1)
        nptest.assert_allclose(data['mean_soil_moisture'][:25],
                               ssm_mean_should,
                               atol=0.1)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import unittest
from datetime import datetime

import numpy as np

import ascat


class TestAscatNcV55R12(unittest.TestCase):

    def setUp(self):

        path = os.path.dirname(__file__)

        cdr_path = os.path.join(path, 'ascat_test_data', 'tuw', 'ascat', 'ssm',
                                '55R12')

        grid_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'grid')

        ioclass_kws = {'loc_dim_name': 'gp', 'loc_ids_name': 'gpi'}

        self.ascat_reader = ascat.AscatSsmCdr(cdr_path, grid_path,
                                              ioclass_kws=ioclass_kws)

    def test_read(self):

        gpi = 2329253
        result = self.ascat_reader.read(gpi)

        assert result.gpi == gpi
        np.testing.assert_approx_equal(result.longitude, 14.28413,
                                       significant=4)
        np.testing.assert_approx_equal(result.latitude, 45.698074,
                                       significant=4)

        assert len(result.data) == 2292
        assert result.data.iloc[15].name == datetime(2007, 1, 15, 19, 34, 41,
                                                     771033)
        assert result.data.iloc[15]['sm'] == 52
        assert result.data.iloc[15]['ssf'] == 1
        assert result.data.iloc[15]['sm_noise'] == 7
        assert result.data.iloc[15]['orbit_dir'].decode('utf-8') == 'A'
        assert result.data.iloc[15]['proc_flag'] == 0

    def test_neighbor_search(self):

        gpi, distance = self.ascat_reader.grid.find_nearest_gpi(3.25, 46.13)
        assert gpi == 2346869
        np.testing.assert_approx_equal(distance, 2267.42, significant=2)


class TestAscatNcV55R21(unittest.TestCase):

    def setUp(self):
        path = os.path.dirname(__file__)

        cdr_path = os.path.join(path, 'ascat_test_data', 'tuw', 'ascat', 'ssm',
                                '55R21')

        grid_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'grid')

        static_layer_path = os.path.join(path, 'ascat_test_data', 'hsaf',
                                         'static_layer')

        self.ascat_reader = ascat.AscatSsmCdr(
            cdr_path, grid_path, static_layer_path=static_layer_path)

    def tearDown(self):
        self.ascat_reader.close()

    def test_read(self):

        gpi = 2329253
        result = self.ascat_reader.read(gpi, absolute_sm=True)
        assert result.gpi == gpi
        np.testing.assert_approx_equal(
            result.longitude, 14.28413, significant=4)
        np.testing.assert_approx_equal(
            result.latitude, 45.698074, significant=4)

        assert len(result.data) == 2457
        assert result.data.iloc[15].name == datetime(
            2007, 1, 15, 19, 34, 41, 5)
        assert result.data.iloc[15]['sm'] == 55
        assert result.data.iloc[15]['ssf'] == 1
        assert result.data.iloc[15]['sm_noise'] == 7
        assert result.data.iloc[15]['frozen_prob'] == 29
        assert result.data.iloc[15]['snow_prob'] == 0
        assert result.data.iloc[15]['orbit_dir'].decode('utf-8') == 'A'
        assert result.data.iloc[15]['proc_flag'] == 0

        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_gldas'], 0.2969999, significant=6)

        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_gldas'], 0.03779999,
            significant=6)

        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_hwsd'], 0.2364999, significant=6)
        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_hwsd'], 0.030100, significant=6)

        assert result.topo_complex == 14
        assert result.wetland_frac == 0

        np.testing.assert_approx_equal(
            result.porosity_gldas, 0.539999, significant=5)
        np.testing.assert_approx_equal(
            result.porosity_hwsd, 0.4299994, significant=5)

    def test_neighbor_search(self):

        gpi, distance = self.ascat_reader.grid.find_nearest_gpi(3.25, 46.13)
        assert gpi == 2346869
        np.testing.assert_approx_equal(distance, 2267.42, significant=2)


class TestAscatNcV55R22(unittest.TestCase):

    def setUp(self):

        path = os.path.dirname(__file__)

        cdr_path = os.path.join(path, 'ascat_test_data', 'tuw', 'ascat', 'ssm',
                                '55R22')

        grid_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'grid')

        static_layer_path = os.path.join(path, 'ascat_test_data', 'hsaf',
                                         'static_layer')

        self.ascat_reader = ascat.AscatSsmCdr(
            cdr_path, grid_path, static_layer_path=static_layer_path)

    def tearDown(self):
        self.ascat_reader.close()

    def test_read(self):

        gpi = 2329253
        result = self.ascat_reader.read(gpi, absolute_sm=True)
        assert result.gpi == gpi
        np.testing.assert_approx_equal(result.longitude, 14.28413,
                                       significant=4)
        np.testing.assert_approx_equal(result.latitude, 45.698074,
                                       significant=4)

        ref_list = ['orbit_dir', 'proc_flag', 'sm', 'sm_noise', 'ssf',
                    'snow_prob', 'frozen_prob', 'abs_sm_gldas',
                    'abs_sm_noise_gldas', 'abs_sm_hwsd', 'abs_sm_noise_hwsd']

        assert sorted(list(result.data.columns)) == sorted(ref_list)

        assert len(result.data) == 2642
        assert result.data.iloc[15].name == datetime(
            2007, 1, 15, 19, 34, 41, 5)
        assert result.data.iloc[15]['sm'] == 55
        assert result.data.iloc[15]['ssf'] == 1
        assert result.data.iloc[15]['sm_noise'] == 7
        assert result.data.iloc[15]['frozen_prob'] == 29
        assert result.data.iloc[15]['snow_prob'] == 0
        assert result.data.iloc[15]['orbit_dir'].decode('utf-8') == 'A'
        assert result.data.iloc[15]['proc_flag'] == 0

        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_gldas'], 0.2969999, significant=6)
        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_gldas'], 0.03779999,
            significant=6)

        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_hwsd'], 0.2364999, significant=6)
        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_hwsd'], 0.0301000, significant=6)

        assert result.topo_complex == 14
        assert result.wetland_frac == 0

        np.testing.assert_approx_equal(
            result.porosity_gldas, 0.539999, significant=5)
        np.testing.assert_approx_equal(
            result.porosity_hwsd, 0.4299994, significant=5)

    def test_neighbor_search(self):

        gpi, distance = self.ascat_reader.grid.find_nearest_gpi(3.25, 46.13)
        assert gpi == 2346869
        np.testing.assert_approx_equal(distance, 2267.42, significant=2)


if __name__ == '__main__':
    unittest.main()
